/*************************************************/
class EvaBimodal{
  int n;
  bimodal **b;
public:
  EvaBimodal();
  ~EvaBimodal();
  void step(int, uint64_t);
};

EvaBimodal::EvaBimodal(){
  n=9;
  b = new bimodal*[n];
  for(int i=0; i<n; i++)
    b[i] = new bimodal(12+i);
}

EvaBimodal::~EvaBimodal(){
  for(int i=0; i<n; i++) delete b[i];
}

void EvaBimodal::step(int r, uint64_t pc){
  int p; /* predicted value */
  for(int i=0; i<n; i++){
    b[i]->predict(pc, &p);
    b[i]->regist(pc, r, p);
  }
}
/*************************************************/
/************************************************/
class EvaGshare{
  gshare **b12;
  gshare **b13;
  gshare **b14;
  gshare **b15;
  gshare **b16;
  gshare **b17;
  gshare **b18;
  gshare **b19;
  gshare **b20;
public:
  EvaGshare();
  ~EvaGshare();
  void step(int,uint64_t,uint64_t);
};

EvaGshare::EvaGshare(){
  b12 = new gshare*[12];
  b13 = new gshare*[13];
  b14 = new gshare*[14];
  b15 = new gshare*[15];
  b16 = new gshare*[16];
  b17 = new gshare*[17];
  b18 = new gshare*[18];
  b19 = new gshare*[19];
  b20 = new gshare*[20];
  for(int i=0; i<12; i++) b12[i] = new gshare(12, i+1);
  for(int i=0; i<13; i++) b13[i] = new gshare(13, i+1);
  for(int i=0; i<14; i++) b14[i] = new gshare(14, i+1);
  for(int i=0; i<15; i++) b15[i] = new gshare(15, i+1);
  for(int i=0; i<16; i++) b16[i] = new gshare(16, i+1);
  for(int i=0; i<17; i++) b17[i] = new gshare(17, i+1);
  for(int i=0; i<18; i++) b18[i] = new gshare(18, i+1);
  for(int i=0; i<19; i++) b19[i] = new gshare(19, i+1);
  for(int i=0; i<20; i++) b20[i] = new gshare(20, i+1);
}

EvaGshare::~EvaGshare(){
  for(int i=0; i<12; i++) delete b12[i];
  for(int i=0; i<13; i++) delete b13[i];
  for(int i=0; i<14; i++) delete b14[i];
  for(int i=0; i<15; i++) delete b15[i];
  for(int i=0; i<16; i++) delete b16[i];
  for(int i=0; i<17; i++) delete b17[i];
  for(int i=0; i<18; i++) delete b18[i];
  for(int i=0; i<19; i++) delete b19[i];
  for(int i=0; i<20; i++) delete b20[i];
}

void EvaGshare::step(int r, 
		     uint64_t pc, uint64_t bhr){
  int p; /* predicted value */
  for(int i=0; i<12; i++){ 
    b12[i]->predict(pc, bhr, &p);
    b12[i]->regist(pc, bhr, r, p);
  }
  for(int i=0; i<13; i++){
    b13[i]->predict(pc, bhr, &p);
    b13[i]->regist(pc, bhr, r, p);
  }
  for(int i=0; i<14; i++){
    b14[i]->predict(pc, bhr, &p);
    b14[i]->regist(pc, bhr, r, p);
  }
  for(int i=0; i<15; i++){
    b15[i]->predict(pc, bhr, &p);
    b15[i]->regist(pc, bhr, r, p);
  }
  for(int i=0; i<16; i++){
    b16[i]->predict(pc, bhr, &p);
    b16[i]->regist(pc, bhr, r, p);
  }
  for(int i=0; i<17; i++){
    b17[i]->predict(pc, bhr, &p);
    b17[i]->regist(pc, bhr, r, p);
  }
  for(int i=0; i<18; i++){
    b18[i]->predict(pc, bhr, &p);
    b18[i]->regist(pc, bhr, r, p);
  }
  for(int i=0; i<19; i++){
    b19[i]->predict(pc, bhr, &p);
    b19[i]->regist(pc, bhr, r, p);
  }
  for(int i=0; i<20; i++){
    b20[i]->predict(pc, bhr, &p);
    b20[i]->regist(pc, bhr, r, p);
  }
}
/*************************************************/
/************************************************/
class EvaBimode{
  bimode **b11;
  bimode **b12;
  bimode **b13;
  bimode **b14;
  bimode **b15;
  bimode **b16;
  bimode **b17;
  bimode **b18;
  bimode **b19;
public:
  EvaBimode();
  ~EvaBimode();
  void step(int,uint64_t,uint64_t);
};

EvaBimode::EvaBimode(){
  b11 = new bimode*[11];
  b12 = new bimode*[12];
  b13 = new bimode*[13];
  b14 = new bimode*[14];
  b15 = new bimode*[15];
  b16 = new bimode*[16];
  b17 = new bimode*[17];
  b18 = new bimode*[18];
  b19 = new bimode*[19];
  for(int i=0; i<11; i++) b11[i] = new bimode(11, i+1, 11);
  for(int i=0; i<12; i++) b12[i] = new bimode(12, i+1, 12);
  for(int i=0; i<13; i++) b13[i] = new bimode(13, i+1, 13);
  for(int i=0; i<14; i++) b14[i] = new bimode(14, i+1, 14);
  for(int i=0; i<15; i++) b15[i] = new bimode(15, i+1, 15);
  for(int i=0; i<16; i++) b16[i] = new bimode(16, i+1, 16);
  for(int i=0; i<17; i++) b17[i] = new bimode(17, i+1, 17);
  for(int i=0; i<18; i++) b18[i] = new bimode(18, i+1, 18);
  for(int i=0; i<19; i++) b19[i] = new bimode(19, i+1, 19);
}

EvaBimode::~EvaBimode(){
  for(int i=0; i<11; i++) delete b11[i];
  for(int i=0; i<12; i++) delete b12[i];
  for(int i=0; i<13; i++) delete b13[i];
  for(int i=0; i<14; i++) delete b14[i];
  for(int i=0; i<15; i++) delete b15[i];
  for(int i=0; i<16; i++) delete b16[i];
  for(int i=0; i<17; i++) delete b17[i];
  for(int i=0; i<18; i++) delete b18[i];
  for(int i=0; i<19; i++) delete b19[i];
}

void EvaBimode::step(int r, uint64_t pc, 
		     uint64_t bhr){
  int p; /* predicted value */
  for(int i=0; i<11; i++){
    bimode *b = b11[i];
    b->predict(pc, bhr, &p);
    b->regist(pc, bhr, r, p);
  }
  for(int i=0; i<12; i++){
    bimode *b = b12[i];
    b->predict(pc, bhr, &p);
    b->regist(pc, bhr, r, p);
  }
  for(int i=0; i<13; i++){
    bimode *b = b13[i];
    b->predict(pc, bhr, &p);
    b->regist(pc, bhr, r, p);
  }
  for(int i=0; i<14; i++){
    bimode *b = b14[i];
    b->predict(pc, bhr, &p);
    b->regist(pc, bhr, r, p);
  }
  for(int i=0; i<15; i++){
    bimode *b = b15[i];
    b->predict(pc, bhr, &p);
    b->regist(pc, bhr, r, p);
  }
  for(int i=0; i<16; i++){
    bimode *b = b16[i];
    b->predict(pc, bhr, &p);
    b->regist(pc, bhr, r, p);
  }
  for(int i=0; i<17; i++){
    bimode *b = b17[i];
    b->predict(pc, bhr, &p);
    b->regist(pc, bhr, r, p);
  }
  for(int i=0; i<18; i++){
    bimode *b = b18[i];
    b->predict(pc, bhr, &p);
    b->regist(pc, bhr, r, p);
  }
  for(int i=0; i<19; i++){
    bimode *b = b19[i];
    b->predict(pc, bhr, &p);
    b->regist(pc, bhr, r, p);
  }
}
/*************************************************/
/************************************************/
void branch_predict(uint64_t npc, uint64_t tpc){
  static uint64_t bhr=0; /* branch history reg */
  uint64_t pc = npc - 4; /* PC of the inst     */
  int result = npc!=tpc; /* branch taken ?     */
  
  static class EvaBimodal eva_bimodal;
  eva_bimodal.step(result, pc);
  
  static class EvaGshare eva_gshare;
  eva_gshare.step(result, pc, bhr);

  static class EvaBimode eva_bimode;
  eva_bimode.step(result, pc, bhr);

  bhr = (bhr << 1) | result;  /* Update BHR */
}
/************************************************/
