/**** Yet Another Alpha Simulator: SimAlpha        by Kenji KISE ****/
/********************************************************************/
#include "define.h"

inline void house_keeper(system_manager *sys, system_config *ms, 
			 evaluation_result *e, debug *deb){
  if(ms->debug_flag) deb->debugmode();
  if(ms->verbose_interval && 
     (e->retired_inst % ms->verbose_interval)==0){
    fprintf(stdout, "."); fflush(stdout);
  }
  if(ms->end_tc!=0 && e->retired_inst>=ms->end_tc) 
    sys->running = 0;
}


simple_chip::simple_chip(char *prog, char **opt){
  sc  = new system_config(prog, opt);
  e   = new evaluation_result;
  as  = new architecture_state(sc, e);
  mem = new memory_system(sc, e);
  deb = new debug(as, mem, sc, e);
  sys = new system_manager(as, mem, sc, e);
  p   = new instruction(as, mem, sys, sc, e);
}


simple_chip::~simple_chip(){   /** destructor **/
  print_evaluation_result(e, sc, sys);
  delete p;
  delete sys;
  delete deb;
  delete mem;
  delete as;
  delete e;
  delete sc;
}


inline int execute_cmovb(instruction *p, architecture_state *as){
  INST_TYPE ir=cmov_ir_create(p->get_ir()); 
  if(ir){ /** CMOVb inst, see README.txt **/
    p->Fetch(&as->pc, ir);
    p->Slot();
    p->Rename();
    p->Issue();
    p->RegisterRead();
    p->Execute(&as->pc);
    p->Memory();
    p->WriteBack();  
  }
  return 0;
}


int simple_chip::step(){
  p->Fetch(&as->pc);    /* pipeline stage 0 */
  p->Slot();            /* pipeline stage 1 */
  p->Rename();          /* pipeline stage 2 */
  p->Issue();           /* pipeline stage 3 */
  p->RegisterRead();    /* pipeline stage 4 */
  p->Execute(&as->pc);  /* pipeline stage 5 */
  p->Memory();          /* pipeline stage 6 */
  p->WriteBack();

  /* split a conditional move,see README.txt */
  execute_cmovb(p, as);

  e->retired_inst++;
  house_keeper(sys, sc, e, deb);

  return sys->running;
}
